#!/usr/bin/env python
#
#   XenMan   -  Copyright (c) 2006 Yves Perrenoud, Jd & Hap Hazard
#   ======
#
# XenMan is a Xen management tool with a GTK based graphical interface
# that allows for performing the standard set of domain operations
# (start, stop, pause, kill, shutdown, reboot, snapshot, etc...). It
# also attempts to simplify certain aspects such as the creation of
# domains, as well as making the consoles available directly within the
# tool's user interface.
#
#
# This software is subject to the GNU Lesser General Public License (LGPL)
# and for details, please consult it at:
#
#    http://www.fsf.org/licensing/licenses/lgpl.txt
#

import gtk, gtk.glade, gobject
from xen.xend.XendClient import server as defaultServer
from xen.xend.XendProtocol import XendError
import xen.xm.create
import vte, re, os, sys, rpm, platform, subprocess
import ConfigParser
import shutil, urllib
import socket
import xmlrpclib
from xen.util.xmlrpclib2 import ServerProxy


#
# constants definitions
#

prop_disks_dir = 'disks_dir'
prop_snapshots_dir = 'snapshots_dir'
prop_xenconf_dir = 'xenconf_dir'
prop_cache_dir = 'cache_dir'
prop_kernel = 'kernel'
prop_ramdisk = 'ramdisk'
prop_dom0_kernel = 'dom0_kernel'
prop_staging_location = 'staging_location'
prop_staging_path_kernel = 'staging_path_kernel'
prop_staging_path_ramdisk = 'staging_path_ramdisk'
prop_snapshot_file_ext = 'snapshot_file_ext'
prop_bootloader = 'bootloader'
prop_lvm = 'use_lvm'
prop_domfiles = "doms"

if re.search('64',platform.architecture()[0]):
    prop_arch_lib_dir = 'lib64'
else:
    prop_arch_lib_dir = 'lib'


#
# defaults
#

default_tcp_port = "8005"
default_path = "/RPC2"
default_user = "root"


for path in (os.path.dirname(sys.argv[0]),'.', '/usr/share/xenman', '.'):
    filename = path + '/xenman.glade'
    if os.path.exists(filename):
        gladefile = filename
        break
else:
    print "ERROR: Couldn't find glade interface definition file!"
    sys.exit(1)


paused_pb   = gtk.gdk.pixbuf_new_from_file(path +\
                                           "/pixmaps/small_pause.png")
resident_pb = gtk.gdk.pixbuf_new_from_file(path + \
                                           "/pixmaps/small_start.png")
not_resident_pb = gtk.gdk.pixbuf_new_from_file(path + \
                                               "/pixmaps/small_shutdown.png")



#
# class definitions
#

class Dom:
    """This base class represents a dom"""

    def isDom0(self):
        """Test wether the dom is dom0"""

        return isinstance(self, DomMem) and self.id == 0

    def isDomU(self):
        """Test wether the dom is a guest dom"""

        return isinstance(self, DomMem) and self.id > 0

class DomMem(Dom):
    """This class represents a dom resident in physical memory"""

    isResident = True
    file = None

    def __init__(self, dom):
        try:
            self.id = int(dom)
            self.name = self["name"]
        except ValueError:
            self.name = dom
            self.id = int(self["domid"])

    def __getitem__(self, param):

        if hasattr(self, "id"):
            dom = self.id
        else:
            dom = self.name
        
        return self.searchtree(server.xend.domain(dom)[1:], param)

    def save(self, filename):
        """Save a snapshot of the dom"""

        server.xend.domain.save(self.id, filename)

    def execShutdownCmd(self, cmd):
        """Execute the requested shutdown command on the dom"""

        server.xend.domain.shutdown(self.id, cmd)

    def reboot(self):
        """Reboot the dom"""

        self.execShutdownCmd("reboot")

    def shutdown(self):
        """Shutdown the dom"""

        self.execShutdownCmd("halt")

    def destroy(self):
        """Destroy the dom, hence removing it from resident running doms
        without any attempt at shutting anything down on the guest OS"""

        server.xend.domain.destroy(self.id)

    def pause(self):
        """Pause the dom"""

        server.xend.domain.pause(self.id)
        
    def resume(self):
        """Resume the paused dom"""

        server.xend.domain.unpause(self.id)

    def setMem(self, value):
        """Change the running memory setting"""

        server.xend.domain.setMemoryTarget(self.id, value)

    def setVCPUs(self, value):
        """Change the running number of CPUs setting"""

        server.xend.domain.setVCpuCount(self.id, value)
        
    @classmethod
    def searchtree(cls, tree, key):
        """Retrieve a value from a tree"""

        for elem in tree:
            if elem[0] == key:
                return elem[1]
            elif type(elem[1]) is list:
                value = cls.searchtree(elem[1], key)
                if value:
                    return value
        return None

class DomFile(Dom):
    """This class represents a domain config file on disk"""

    class DiskEntry:
        """This class provides a structured representation of the disk
        string. The following attributes are available:

         type:     {file|lvm|part}
         filename: the path to the VBD or device
         device:   the device to be used in the guest system
         mode:     {r|w}
         """

        def __init__(self, input):
            if type(input) is str:
                m = re.match("^(phy|file):(\S+),([a-z|0-9]+|ioemu:[a-z|0-9]+),(r|w)$", input)
                if m:
                    self.filename = m.group(2)
                    self.device = m.group(3)
                    self.mode = m.group(4)

                    if m.group(1) == "file":
                        self.type = "file"
                    else:
                        device = os.path.realpath(m.group(2))
                        if device.startswith("/dev/mapper"):
                            self.type = "lvm"
                        else:
                            self.type = "part"
            elif type(input) in (list, tuple):
                self.type, self.filename, self.device, self.mode = input

        def __repr__(self):
            if self.type == "file":
                dtype = "file"
            else:
                dtype = "phy"

            return "%s:%s,%s,%s" % (dtype, self.filename, self.device,
                                    self.mode)

    isResident = False

    def __init__(self, filename):
        self.options = {}
        self.filename = filename
        
        known_options = set(['kernel','ramdisk', 'vif', 'disk',
                             'device_model', 'builder', 'name',
                             'vcpus','pae','acpi','apic','cpus',
                             'on_poweroff','on_reboot','on_crash',
                             'on_shutdown','cdrom','boot','snapshot','sdl',
                             'vnc','vncviewer','nographic','stdvga',
                             'serial','ne2000','audio','localtime',
                             'full-screen','memory','bootloader',
                             'extra','root'])
        imports = ["import sys",
                   "import os",
                   "import os.path",
                   "from xen.util.ip import *"]
        cmd = '\n'.join(imports + ["from xen.xm.help import Vars",
                         "xm_file = '%s'" % filename,
                         "xm_help = %d" % 0,
                         "xm_vars = Vars(xm_file, xm_help, locals())"])


        if os.path.exists(filename):
            globs = {}
            locs = {}
            exec cmd in globs, locs
            #try:
            f = open(filename)
            lines = f.readlines()
            f.close()
            cmd = "\n".join(lines)
            exec cmd in globs, locs            
            #except:
            #    raise
                    
            for name in locs:
                if name in known_options:
                    self.options[name] = locs[name]
        
            self.name = self["name"]

        for opt in ("disk", "vif"):
            if not self.options.has_key(opt):
                self.options[opt] = []


    def __getitem__(self, name):
        if self.options.has_key(name):
            return self.options[name]
        else:
            attrib = getattr(self,name, None)
            if attrib is not None:
                return attrib
            else:
                return None

    def __setitem__(self, name, item):
        self.options[name] = item

    def __iter__(self):
        return self.options.iterkeys()

    def getDisks(self):
        """This method returns a more structured version of the config file's
        disk entry"""

        reslist = []
        for str in self["disk"]:
            reslist.append(self.DiskEntry(str))
        return reslist

    def write(self):
        """Writes the settings out to the filename specified during
        initialization"""
        
        self.name = self["name"]
        if os.path.exists(self.filename):
            shutil.copy(self.filename, self.filename + '.bak')
        outfile = open(self.filename, 'w')
        outfile.write("# Automtically generated by XenMan\n")

        for name, value in self.options.iteritems():
            outfile.write("%s = %s\n" % (name, repr(value)))

    def validate(self):
        """Attempts to validate that the settings are not going to lead to
        any errors when the dom is started, and returns a list of the
        errors as strings"""

        result = []

        if not self["name"]:
            result.append("Missing domain name.")

        if not self["disk"]:
            result.append("Missing disk specification.")

        if not self["bootloader"]:
            # if a bootloader is not specified,
            # check for a valid kernel            
            if not self['kernel'] or not os.path.exists(self['kernel']):
                result.append("Invalid file name for %s." % 'kernel')
            if self['ramdisk'] and not os.path.exists(self['ramdisk']):
                result.append("Invalid file name for %s." % 'ramdisk')
                    

        for parm, nodeval in (("memory", "total_memory"),
                               ("vcpus", "nr_cpus")):
            if self[parm] and int(self[parm]) > int(getNodeVal(nodeval)):
                result.append("More %s requested than total %s." % (parm,
                                                                    parm))

        return result

    def safe_start(self):
        r, w = os.pipe() # these are file descriptors, not file objects
        
        pid = os.fork()
        if pid:
            # we are the parent
            os.close(w) #  close a writer file descriptor
            r = os.fdopen(r) # turn r into a file object
            txt = ""
            txt = r.read()
            r.close()
            os.waitpid(pid, 0) # wait for child process gets cleaned up
            if txt.find("_xenman_success_") < 0:
                raise OSError("xm create failed." + txt)
        else:
            # we are the child
            os.close(r)
            wf = os.fdopen(w, 'w')
            # dup stdout and std error to end of the pipe
            os.dup2(w,sys.stdout.fileno())  
            os.dup2(w,sys.stderr.fileno())
            # call the actual create
            # TODO: error handling.. and pass message to parent.
            try:
                xen.xm.create.main(("create", self.filename))
                wf.write("_xenman_success_")  # Simple IPC for success
            except Exception,ex:
                wf.write(str(ex))
            except Error, err:
                wf.write(str(err))
                
            wf.close()
            sys.exit(0)
    

    def start(self):
        """Start the dom"""
        #xen.xm.create.main(("create", self.filename))
        #self.safe_start()
        p1 = subprocess.Popen('/usr/sbin/xm create %s'% self.filename,\
                              stderr=subprocess.PIPE, \
                              universal_newlines=True,\
                              shell=True, close_fds=True)
        error = p1.communicate()[1]
        if (p1.returncode):
            raise OSError(error)
        else:
            newdom = DomMem(self.name)
            newdom.file = self
            return newdom
        

class DomIndex:
    """This class manages the DomView gtk.ListStore and it's instance can be
    used to access Dom objects for each element in the list"""

    xenauto = "/etc/xen/auto"

    def __init__(self, wtree):
        self.wtree = wtree
        self.domview = wtree.get_widget("DomView")
        self.domview.get_selection().connect("changed", self.domselected)
        
        pbrenderer = gtk.CellRendererPixbuf()
        column = gtk.TreeViewColumn("State", pbrenderer)
        column.set_cell_data_func(pbrenderer, self.get_state_pixbuf)
        self.domview.append_column(column)

        textrenderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn("Name", textrenderer, text=0)
        self.domview.append_column(column)

        self.domlistmodel = gtk.ListStore(gobject.TYPE_STRING)
        slistmodel = gtk.TreeModelSort(self.domlistmodel)
        slistmodel.set_sort_column_id(0, gtk.SORT_ASCENDING)
        self.domview.set_model(slistmodel)

        self.notebook = NoteBook(self)

        self.domdict = {}
        self.extradoms = config.getDefault(prop_domfiles)
        if self.extradoms is None:
            self.extradoms = []
        else:
            self.extradoms = eval(self.extradoms)
        self.userpause = True
        self.refresh()

    def __getitem__(self, item):
        if type(item) is int:
            for name, dom in self.domdict.iteritems():
                if dom.isResident and dom.id == item:
                    return dom
        else:
            return self.domdict[item]

    def __iter__(self):
        return self.domdict.iterkeys()

    def getDomains(self):
        """ return list of domain names """
        return server.xend.domains(0)

    def listResident(self):
        """Return a list of doms that are resident in physical memory"""

        domlist = []
        for dom in self.domdict.itervalues():
            if dom.isResident:
                domlist.append(dom)
        
        return domlist

    def addDomFile(self, filename):
        """Adds a dom specified by it's file name to the index"""

        if filename.startswith(self.xenauto) or filename in self.extradoms:
            return False
        else:
            try: 
                dom = DomFile(filename)
                self.extradoms.append(filename)
                config.setDefault(prop_domfiles, repr(self.extradoms))
                
                if hasattr(self, "cursel") and self.cursel.isResident \
                       and self.cursel.name == dom.name:
                    self.cursel.file = dom
                    
                self.refresh()
            except TypeError, err:
                showmsg("\nFAILED: Couldn't parse configuration\n\n" +
                        "Please consult the documentation for more "+
                        "information on supported file types\n\n" +
                        "Error: "+ str(err))
            except Exception, exc:
                showmsg("\nFAILED: Couldn't parse configuration\n\n" +
                        "Error: "+
                        str(exc))

    def remDomFile(self, filename):
        """Remove a dom specified by it's file name to the index"""

        if filename in self.extradoms:
            self.extradoms.remove(filename)
            config.setDefault(prop_domfiles, repr(self.extradoms))
            self.refresh()
        else:
            return False

    def refresh(self):
        """Refresh DomView with a current list of Doms"""

        curdomdict = {}
        try:
            for dom in self.getDomains():
                try:
                    curdomdict[dom] = DomMem(dom)
                except (XendError, xmlrpclib.Fault):
                    continue
        except (socket.error, xmlrpclib.Fault):
            return

        for filename in os.listdir(self.xenauto) + self.extradoms:
            try:
                if filename.find('/') >= 0:
                    if os.path.exists(filename):
                        dom = DomFile(filename)
                    else:
                        dom = DomFile("%s/%s" % (self.xenauto, filename))
            except :
                if filename in self.extradoms: self.extradoms.remove(filename)
                print "Couldn't read %s. Removing from the list" % filename
                continue
            
            domname = dom.name
            if domname not in curdomdict:
                curdomdict[domname] = dom
            else:
                curdomdict[domname].file = dom

        oldset = set(self.domdict.keys())
        newset = set(curdomdict.keys())

        for dom in oldset - newset:
            self.domlistmodel.remove(self.domdict[dom].iter)
            del(self.domdict[dom])

        for dom in newset - oldset:
            self.domdict[dom] = curdomdict[dom]
            self.domdict[dom].iter = self.domlistmodel.append([dom])

        changes = 0
        for dom in oldset & newset:
            if (isinstance(curdomdict[dom], DomFile) and not \
                isinstance(self.domdict[dom], DomFile)) or \
                (isinstance(curdomdict[dom], DomMem) and not \
                 isinstance(self.domdict[dom], DomMem)):
                iter = self.domdict[dom].iter
                self.domdict[dom] = curdomdict[dom]
                self.domdict[dom].iter = iter
                changes += 1

        if changes:
            loop.uirefresh = True

        self.notebook.cleanup()

    def domselected(self, widget=None):
        """Handles the selection of a domain in the domain tree view"""

        if not widget:
            widget = self.domview.get_selection()

        treemodel, treeiter = widget.get_selected()

        if treeiter:
            self.cursel = self[treemodel.get_value(treeiter, 0)]

            if not self.cursel.isDomU():
                if isinstance(self.cursel, DomFile):
                    self.notebook.showInfoTab()
                else:
                   self.notebook.showTab(0)
            else:
                self.notebook.showTab(self.cursel)

            self.stateRefresh()

    def stateRefresh(self, checkiter=False):
        """Refreshes UI elements based on current dom selection"""

        if checkiter:
            widget = self.domview.get_selection()

            treemodel, treeiter = widget.get_selected()

            if treeiter:
                self.cursel = self[treemodel.get_value(treeiter, 0)]

        pausebutton = self.wtree.get_widget("pausebutton")
        pausemenu  = self.wtree.get_widget("menu_pause")
        pausepopup = self.wtree.get_widget("Popup_pause")

        remove = ("MenuRemove", "PopupRemove")
        
        for widget in ("MenuSettings", "PopupSettings"):
            self.wtree.get_widget(widget).set_sensitive(True)

        if self.cursel.isDom0():
            for action in ("start", "pause", "reboot", "shutdown", "kill",
                           "snapshot"):
                self.wtree.get_widget(action + "button").set_sensitive(False)
                self.wtree.get_widget("menu_" + action).set_sensitive(False)
                self.wtree.get_widget("Popup_" + action).set_sensitive(False)

            self.userpause = False
            pausebutton.set_active(False)
            self.userpause = True

            pausemenu.set_sensitive(False)
            pausepopup.set_sensitive(False)

            for widget in remove:
                self.wtree.get_widget(widget).set_sensitive(False)

            return
        else:
            for widget in ("startbutton", "menu_start", "Popup_start"):
                self.wtree.get_widget(widget)\
                                   .set_sensitive(not self.cursel.isResident)

            for action in ("pause", "reboot", "shutdown", "kill", "snapshot"):
                self.wtree.get_widget(action + "button")\
                                    .set_sensitive(self.cursel.isResident)
                self.wtree.get_widget("menu_" + action)\
                                     .set_sensitive(self.cursel.isResident)
                self.wtree.get_widget("Popup_" + action)\
                                     .set_sensitive(self.cursel.isResident)

            isauto = False
            if isinstance(self.cursel, DomMem):
                if self.cursel.file and \
                       self.cursel.file.filename.startswith(self.xenauto):
                    isauto = True
            elif self.cursel.filename.startswith(self.xenauto):
                isauto = True

            for widget in remove:
                if isauto or self.cursel.isResident:
                    self.wtree.get_widget(widget).set_sensitive(False)
                else:
                    self.wtree.get_widget(widget).set_sensitive(True)
            

            if self.cursel.isResident:
                state = self.cursel["state"]
            else:
                state = None

            self.userpause = False
            if state:
                pausemenu.set_sensitive(True)
                pausepopup.set_sensitive(True)

                if (state[2] == 'p'):
                    pausebutton.set_active(True)
                    self.notebook.pausedTab(self.cursel, True)
                    
                    for widget in (pausebutton, pausemenu.child,
                                   pausepopup.child):
                        widget.set_label("Resume")
                else:
                    pausebutton.set_active(False)
                    self.notebook.pausedTab(self.cursel, False)
                    
                    for widget in (pausebutton, pausemenu.child,
                                   pausepopup.child):
                        widget.set_label("Pause")
            else:
                pausebutton.set_active(False)
                for widget in (pausemenu.child, pausepopup.child):
                    widget.set_sensitive(False)
            self.userpause = True
            
    def get_state_pixbuf(self, column, renderer, model, iter):
        """ Call back to display the state column image"""
        dname =  model.get_value(iter,0)
        if domindex[dname].isResident:
            try:
                state = domindex[dname]["state"]
            except xmlrpclib.Fault :
                state = None
                
            if state is not None and state[2] == 'p':
                pb = paused_pb
            else:
                pb = resident_pb
        else:
            pb = not_resident_pb
        renderer.set_property('pixbuf', pb)

class NoteBook:
    """This class manages the gtk.NoteBook"""

    def __init__(self, domindex):
        self.notebook = domindex.wtree.get_widget("MainNotebook")
        self.domindex = domindex
        self.nbooklist = [0,0]
        self.terms = {}
        self.pausedterms = {}
        
        self.init_info_tree()
        # add handlers : connecting through glade not working.
        self.notebook.connect("switch_page", \
                              self.on_notebook_switch_page)

    def __getDomID(self, dom):
        """Always returns the DomID whether the argument in is a int or a dom
        object"""

        if isinstance(dom, Dom):
            domid = dom.id
        else:
            domid = dom

        return domid

    def showInfoTab(self,widget=None,dom=None):
        if self.notebook.get_current_page() == 1:
            self.populateInfoTree()
        else:
            self.notebook.set_current_page(1)
        
            
    def showTab(self, dom):
        """Displays the tab for the requested Dom or Dom ID"""

        domid = self.__getDomID(dom)

        if domid in self.nbooklist:
            self.notebook.set_current_page(self.nbooklist.index(domid))
        else:
            self.__addTab(dom)
            self.pausedTab(domid, False)

    def removeTab(self, domid):
        """Removes the tab for the requested Dom ID"""

        if domid in self.nbooklist:
            del self.terms[domid]
            self.notebook.remove_page(self.nbooklist.index(domid))
            self.nbooklist.remove(domid)
        else:
            raise "No such tab to remove"

    def getTab(self):
        """Returns the Dom ID of the currently selected tab"""

        return self.nbooklist[self.notebook.get_current_page()]

    def pausedTab(self, dom, state):
        """Set the sensitive state of the contents of a tab specified by
        the requested Dom or Dom ID"""

        domid = self.__getDomID(dom)

        if domid in self.nbooklist:
            if state:
                if domid not in self.pausedterms:
                    tbox = self.terms[domid]
                    term = tbox.get_children()[0]
                    self.pausedterms[domid] = term
                    tbox.remove(term)
                    label = gtk.Label()
                    label.set_markup('<span size="xx-large">Paused</span>')
                    tbox.add(label)
                    tbox.show_all()
            elif domid in self.pausedterms:
                term = self.pausedterms[domid]
                del self.pausedterms[domid]
                tbox = self.terms[domid]
                tbox.remove(tbox.get_children()[0])
                tbox.add(term)
        else:
            raise "domid doesn't correspond to a tab"

    def __addTab(self, dom):
        """Adds a console tab to the notebook for the specified Dom"""

        if isinstance(dom, Dom):
            domid = dom.id
            domname = dom.name
        else:
            domid = dom
            domname = self.domindex[domid].name

        term = vte.Terminal()
        term.fork_command("/usr/"+ prop_arch_lib_dir +
                          "/xen/bin/xenconsole",
                          ("xenconsole", str(domid)))
        tbox = gtk.VBox()
        tbox.add(term)
        tbox.show_all()

        nbox = gtk.HBox()
        nbox.pack_start(gtk.Label(domname))
        closebutton = gtk.Button()
        closebutton.set_relief(gtk.RELIEF_NONE)
        closebutton.set_size_request(20, 17)
        closeimage = gtk.Image()
        closeimage.set_from_stock(gtk.STOCK_CLOSE,
                                  gtk.ICON_SIZE_MENU)
        closebutton.add(closeimage)
        closebutton.connect("clicked", self.__closeTab, domid)

        """ Some other time...
        infobutton = gtk.Button()
        infobutton.set_relief(gtk.RELIEF_NONE)
        infobutton.set_size_request(28, 28)
        info_image = gtk.Image()
        info_image.set_from_stock(gtk.STOCK_INFO,
                                  gtk.ICON_SIZE_MENU)
        infobutton.add(info_image)
        infobutton.connect("clicked", self.showInfoTab, dom)
        nbox.pack_start(infobutton)
        """
        nbox.pack_start(closebutton)
        nbox.show_all()

        pageid = self.notebook.append_page(tbox, nbox)
        self.notebook.set_current_page(pageid)
        self.nbooklist.append(domid)
        self.terms[domid] = tbox
        self.domindex.stateRefresh()

    def __closeTab(self, widget, domid):
        """Handles the request to close a notebook tab"""

        self.removeTab(domid)

    def cleanup(self):
        """Goes through the list of notebook tabs and checks if there is a
        corresponding Dom and what it's state is, in order to cleanup tabs
        that no longer should be displayed"""

        changes = 0

        for domid in self.nbooklist:
            try:
                server.xend.domain(domid)
            except (XendError,xmlrpclib.Fault):
                self.removeTab(domid)
                changes += 1
                continue

        if changes:
            # create new tab if required (after reboot)
            self.domindex.domselected()

            
    def on_notebook_switch_page(self, widget, page, page_no):
        """if the user switched to info tab, populate with the
        dom information """
        if page_no == 1:
            self.populateInfoTree()
            
        return True

    def init_info_tree(self):
        self.info_tree_model = gtk.TreeStore(*[gobject.TYPE_STRING]*2)
        self.info_tree_view = wtree.get_widget("info_tree")

        
        textrenderer = gtk.CellRendererText()

        column = gtk.TreeViewColumn("Configuration", textrenderer, text=0)
        self.info_tree_view.append_column(column)

        column = gtk.TreeViewColumn("", textrenderer, text=1)
        self.info_tree_view.append_column(column)

        """
        column = gtk.TreeViewColumn("Value", textrenderer, text=2)
        self.info_tree_view.append_column(column)
        """
        self.info_tree_view.set_model(self.info_tree_model)
        self.info_tree_view.show()


    def populateInfoTree(self):
        selection = getattr(domindex,"cursel",None)
        if selection == None:
            #clean up info tab with "No selection"
            return
        self.info_tree_model.clear()


        
        gen_iter  = self.info_tree_model.append(None,["General",""])
        boot_iter = self.info_tree_model.append(None,["Boot",""])
        res_iter  = self.info_tree_model.append(None,["Resource",""])

        self.populateGeneralInfo(gen_iter)
        self.populateBootInfo(boot_iter)
        self.populateResourceInfo(res_iter)
        self.info_tree_view.expand_all()


    def append_info_node(self, iter, name, param):
        """ get the value from dom index and append Name / Value in the
        tree at 'iter' position."""
        
        value = self.domindex.cursel[param]

        if value:
            if type(value) == int:
                value = str(value)
            self.info_tree_model.append(iter, [name,value])
        else:
            self.info_tree_model.append(iter, [name,"N/A"])

    
            
    def populateGeneralInfo(self, iter):
        for name, param in ( ("Name", "name"),
                             ("FileName", "filename"),
                             ):
            self.append_info_node(iter, name, param)
        

    def populateBootInfo(self, iter):
        for name, param in ( 
                             ("Kernel", "kernel"),
                             ("Ramdisk", "ramdisk"),
                             ("Bootloader", "bootloader"),

                             ("On Crash", "on_crash"),
                             ("On Reboot", "on_reboot"),):
            self.append_info_node(iter, name, param)


    def populateResourceInfo(self, iter):
        for name, param in ( 
                             ("Memory", "memory"),
                             ("CPU", "vcpus"),
                             ("Network", "vif"),
                             ("Disks", "disk"),
                             ):
            self.append_info_node(iter, name, param)

class RefreshLoop:
    """This class initializes a GTK call back loop that runs one of it's
    methods on a regular basis to perform refreshes and such operations"""

    stat_refresh_rate = 5

    def __init__(self, wtree, domindex):
        self.timeoutid = gobject.timeout_add(1000, self.__callback)
        self.domindex = domindex
        self.skipcount = 0
        self.uirefresh = False
        self.stat_refresh_count = self.stat_refresh_rate

        self.statmodel = gtk.ListStore(*[gobject.TYPE_STRING]*5)

        statview = wtree.get_widget("statTable")
        textrenderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn("Id", textrenderer, text=0)
        statview.append_column(column)
        column = gtk.TreeViewColumn("Name", textrenderer, text=1)
        statview.append_column(column)
        column = gtk.TreeViewColumn("State", textrenderer, text=2)
        statview.append_column(column)
        column = gtk.TreeViewColumn("Memory", textrenderer, text=3)
        statview.append_column(column)
        column = gtk.TreeViewColumn("CPU Time(sec)", textrenderer, text=4)
        statview.append_column(column)
        statview.set_model(self.statmodel)
        statview.show()

        self.refreshStats()

    def __callback(self):
        """This is the callback that handles the refresh"""

        self.domindex.refresh()

        self.stat_refresh_count -= 1
        if (self.stat_refresh_count <= 0):
           self.stat_refresh_count = self.stat_refresh_rate
           # fetch stats only if home tab is shown
           if (self.domindex.notebook.getTab() == 0):
               self.refreshStats()

        if self.skipcount:
            self.skipcount -= 1
            return True

        if self.uirefresh:
            self.domindex.stateRefresh(True)
            self.uirefresh = False

        return True

    # For refresh_stats, clearing the model isn't ideal as the selection
    # highlight gets cleared at each refresh, which looks very odd to the
    # user... the append()/remove() method used in DomIndex.refresh() for
    # instance, doesn't suffer that problem - should be fixed

    def refreshStats(self):
        """Refreshes the statistics on the home tab"""

        self.statmodel.clear()

        for dom in domindex.listResident():
            iter = self.statmodel.insert_before(None, None)
            self.statmodel.set(iter,
                               0, dom.id,
                               1, dom.name,
                               2, dom["state"],
                               3, dom["memory"],
                               4, "%8.3f" % float(dom["cpu_time"]),
                               )

class DomSettings:
    """Class that handles everything related to the Dom Settings dialog"""

    class DiskEdit:
        """Class that handles everything related to the disk edit area of
        the settings dialog"""

        diskmapping = ["file", "phy"]
        modemapping = ["w", "r"]
        colmap = ["Type", "Device Path", "Guest Device", "Mode"]
        guestdevs = ["xvda", "xvdb", "xvdc"]

        def __init__(self, wtree, filesel, parentwin):
            self.dialog = wtree.get_widget("DiskEditDialog")
            self.addbutton = wtree.get_widget("DiskAddButton")
            self.propbutton = wtree.get_widget("DiskPropButton")
            self.rembutton = wtree.get_widget("DiskRemoveButton")
            self.disktype = wtree.get_widget("DiskTypeCombo")
            self.diskdev = wtree.get_widget("DiskDeviceEntry")
            self.diskdevfsel = wtree.get_widget("DiskDeviceButton")
            self.guestdev = wtree.get_widget("GuestDevCombo")
            self.mode = wtree.get_widget("ModeCombo")

            self.diskview = wtree.get_widget("DiskTreeView")
            textrenderer = gtk.CellRendererText()

            self.dialog.set_transient_for(parentwin)

            for dev in self.guestdevs:
                self.guestdev.append_text(dev)

            pos = 0
            for colname in self.colmap:
                column = gtk.TreeViewColumn(colname, textrenderer, text=pos)
                self.diskview.append_column(column)
                pos += 1

            wtree.signal_autoconnect(
                {"on_DiskAddButton_clicked": self.__showSettings,
                 "on_DiskPropButton_clicked": (self.__showSettings, True),
                 "on_DiskRemoveButton_clicked": self.__removeClicked,
                 "on_DiskDeviceButton_clicked": (filesel, self.diskdev,
                                                 "open", self.dialog),
                 })

        def __removeClicked(self, widget):
            """Handles the clicking of the remove button in the disk section
            of the settings dialog"""

            model, iter = self.diskview.get_selection().get_selected()

            if iter:
                model.remove(iter)

        def showtree(self, dom=None):
            """Populates the disk tree view with all disks for this dom"""

            self.diskmodel = gtk.ListStore(*[gobject.TYPE_STRING]*4)
            self.diskview.set_model(self.diskmodel)

            if dom:
                for disk in dom.getDisks():
                    if disk.type == "file":
                        dtype = disk.type
                    else:
                        dtype = "phy"

                    self.diskmodel.append([dtype, disk.filename,
                                           disk.device, disk.mode])

        def __showSettings(self, widget, edit=False):
            """Show the dialog and optionally populate with the selected
            disk entry"""

            model, iter = self.diskview.get_selection().get_selected()

            if edit and iter:
                for combo, mapping, entry in (
                    (self.disktype, self.diskmapping, "Type"),
                    (self.mode, self.modemapping, "Mode")):
                    combo.set_active(mapping.index(
                        model.get_value(iter, self.colmap.index(entry))))

                self.diskdev.set_text(
                    model.get_value(iter, self.colmap.index("Device Path")))

                guestdev =  model.get_value(iter,
                                            self.colmap.index("Guest Device"))
                if guestdev not in self.guestdevs:
                    self.guestdev.prepend_text(guestdev)
                    self.guestdev.set_active(0)
                else:
                    self.guestdev.set_active(self.guestdevs.index(guestdev))
            else:
                self.diskdev.set_text('')

                # The following three lines are a hack to compensate for the
                # fact that the following call doesn't work as documented
                ## self.guestdev.set_active(-1)
                self.guestdev.prepend_text('')
                self.guestdev.set_active(0)
                self.guestdev.remove_text(0)

                for combo in (self.disktype, self.mode):
                    combo.set_active(0)

            res = self.dialog.run() == gtk.RESPONSE_OK

            if not res:
                self.dialog.hide()
                return

            if edit and iter:
                model.remove(iter)

            model.append(
                [self.diskmapping[self.disktype.get_active()],
                 self.diskdev.get_text(),
                 self.guestdev.get_active_text(),
                 self.modemapping[self.mode.get_active()],
                 ])

            self.dialog.hide()

        def getDiskEntries(self):
            """Returns a list of DiskEntry instances for each row of the
            ListStore"""

            delist = []
            iter = self.diskmodel.get_iter_first()
            while iter:
                diskparms = []
                for row in enumerate(self.colmap):
                    diskparms.append(self.diskmodel.get_value(iter, row[0]))

                delist.append(DomFile.DiskEntry(diskparms))
                iter = self.diskmodel.iter_next(iter)

            return delist

    combomapping = ["destroy", "restart", "preserve", "rename-restart"]

    textfields = ("name", "kernel", "ramdisk", "root", "extra", "bootloader")
    valuefields = ("memory", "vcpus")
    combofields = ("on_shutdown", "on_reboot", "on_crash")

    minmem = 16
    defaultmem = 256
    ignoreTypeChange = False

    def __init__(self, wtree, domindex):
        self.domindex = domindex
        self.dialog = wtree.get_widget("SettingsDialog")

        self.kernel = wtree.get_widget("KernelEntry")
        self.ramdisk = wtree.get_widget("RamDiskEntry")
        self.use_bootloader = wtree.get_widget("UseBootloaderCheckButton")
        self.bootloader = wtree.get_widget("BootloaderEntry")
        self.rootdev = wtree.get_widget("RootDevEntry")
        self.kernargs = wtree.get_widget("KernArgsEntry")
        self.diskframe = wtree.get_widget("DiskFrame")
        self.usablecpu = wtree.get_widget("UsableCPUEntry")
        self.vcpu = wtree.get_widget("VCPUSpin")
        self.power = wtree.get_widget("PowerCombo")
        self.reboot = wtree.get_widget("RebootCombo")
        self.crash = wtree.get_widget("CrashCombo")
        self.networkframe = wtree.get_widget("NetworkFrame")
        self.name = wtree.get_widget("NameEntry")
        self.kernelsel = wtree.get_widget("KernelButton")
        self.ramdisksel = wtree.get_widget("RamDiskButton")
        self.memory = wtree.get_widget("MemoryScale")
        self.inmemory = wtree.get_widget("InMemory")
        self.ondisk = wtree.get_widget("OnDisk")
        self.filename = wtree.get_widget("FileNameEntry")
        self.filenamesel = wtree.get_widget("FileNameButton")

        self.vcpu.set_range(1.0, float(getNodeVal("nr_cpus")))

        self.diskedit = self.DiskEdit(wtree, self.__filesel, self.dialog)

        self.diskmapping = {"name": self.name, "memory": self.memory,
                            "kernel": self.kernel, "ramdisk": self.ramdisk,
                            "root": self.rootdev, "cpus": self.usablecpu,
                            "extra": self.kernargs, "vcpus": self.vcpu,
                            "on_shutdown": self.power,
                            "on_reboot": self.reboot,
                            "on_crash": self.crash,
                            "bootloader": self.bootloader,}

        wtree.signal_autoconnect(
            {"on_InMemory_clicked": self.__typeChange,
             "on_KernelButton_clicked": (self.__filesel, self.kernel),
             "on_RamDiskButton_clicked": (self.__filesel, self.ramdisk),
             "on_FileNameButton_clicked": (self.__filesel, self.filename,
                                           "save"),
             "on_UseBootloaderCheckButton_toggled": self.__boot_toggled,
             })

    def __boot_toggled(self, widget):
        use_boot = widget.get_active()
        self.kernel.set_sensitive(not use_boot)
        self.ramdisk.set_sensitive(not use_boot)
        self.bootloader.set_sensitive(use_boot)
        
    

    def __typeChange(self, widget):
        """Handles the radio button that allows one to select what kind
        of settings (either memory resident or on disk) one wishes to edit"""

        if self.ignoreTypeChange:
            return

        if self.inmemory.get_active():
            self.type = "memory"
            self.__showMem()
        elif self.ondisk.get_active():
            self.type = "disk"
            self.__showDisk()
        else:
            raise TypeError

    def __noOpActive(self, widget, state):
        """Change the state of the settings type widgets without performing
        any actions as a result"""

        self.ignoreTypeChange = True
        widget.set_active(state)
        self.ignoreTypeChange = False

    def __filesel(self, widget, field, fstype="open", parentwin=None):
        """Handles the pressing of the file selection buttons"""

        if not parentwin:
            parentwin = self.dialog

        filename = field.get_text()

        init_folder = os.path.dirname(filename)
        if fstype == "open":
            init_file = None
        else:
            init_file = filename

        result, filename = file_selection("Select file", fstype, init_folder,
                                          init_file, parentwin)

        if result:
            field.set_text(filename)

    def __dom0WidgetsOn(self, state):
        """sets the state of the widgets that are immutable only for dom0"""

        for widget in (self.diskframe, self.usablecpu, self.vcpu,
                       self.networkframe):
            widget.set_sensitive(state)

    def __diskWidgetsOn(self, state):
        """sets the state of the widgets that only apply to on disk settings"""

        for widget in (self.kernel, self.ramdisk, self.rootdev, self.kernargs,
                       self.kernelsel, self.ramdisksel, self.name,
                       self.usablecpu, self.power, self.reboot, self.crash,
                       self.filename, self.filenamesel, self.diskframe,
                       self.bootloader, self.use_bootloader):
            widget.set_sensitive(state)

    def __showMem(self):
        """Displays the current selection's memory resident settings"""

        self.__dom0WidgetsOn(self.domindex.cursel.id)
        self.__diskWidgetsOn(False)

        if self.domindex.cursel.file:
            self.filename.set_text(self.domindex.cursel.file.filename)

        for widget, parm in ((self.name, "name"),
                             (self.kernel, "kernel"),
                             (self.ramdisk, "ramdisk"),
                             (self.usablecpu, "vcpu_avail"),
                             (self.bootloader, "bootloader"),
                             ):
            value = self.domindex.cursel[parm]
            if value:
                widget.set_text(value)
            else:
                widget.set_text('')

       

        curmem = int(self.domindex.cursel["memory"])
        freemem = int(getNodeVal("free_memory"))
        self.memory.set_range(self.minmem, curmem+freemem)

        for widget, parm in ((self.memory, "memory"), (self.vcpu, "vcpus")):
            widget.set_value(int(self.domindex.cursel[parm]))

        for widget, parm in ((self.power, "on_poweroff"),
                             (self.reboot, "on_reboot"),
                             (self.crash, "on_crash")):
            widget.set_active(
                self.combomapping.index(self.domindex.cursel[parm]))

        for widget in (self.rootdev, self.kernargs):
            widget.set_text('')

        self.diskedit.showtree()

    def __showDisk(self):
        """Displays the current selection's on disk settings"""

        self.__dom0WidgetsOn(True)
        self.__diskWidgetsOn(True)

        if self.domindex.cursel.isResident:
            self.domfile = self.domindex.cursel.file
        else:
            self.domfile = self.domindex.cursel

        self.filename.set_text(self.domfile.filename)

        bl = self.domfile["bootloader"]
        if bl and bl.strip() != '':
            self.use_bootloader.set_active(True)
        else:
            self.use_bootloader.set_active(False)

        self.__boot_toggled(self.use_bootloader)
       
        for entry, widget in self.diskmapping.iteritems():
            if entry in self.textfields:
                if entry in self.domfile:
                    value = self.domfile[entry]
                else:
                    value = ''

                if not value:
                    value = ''
                
                widget.set_text(value)

            elif entry == "memory":
                self.memory.set_range(self.minmem,
                                      int(getNodeVal("total_memory")))

                if entry in self.domfile:
                    value = int(self.domfile[entry])
                else:
                    value = self.defaultmem

                widget.set_value(value)

            elif entry == "vcpus":
                if entry in self.domfile:
                    value = int(self.domfile[entry])
                else:
                    value = 1

                widget.set_value(value)

            elif entry in self.combofields:
                if entry in self.domfile:
                    value = self.combomapping.index(self.domfile[entry])
                else:
                    value = -1

                widget.set_active(value)

        self.diskedit.showtree(self.domfile)

    def __showBlank(self):
        """Clears all the fields"""

        self.__dom0WidgetsOn(True)
        self.__diskWidgetsOn(True)

        self.filename.set_text('')

        for entry in ("name", "kernel", "ramdisk", "root", "extra"):
            self.diskmapping[entry].set_text('')
            
        self.memory.set_value(self.defaultmem)
        self.vcpu.set_value(1)

        for entry in ("on_shutdown", "on_crash"):
            self.diskmapping[entry].set_active(0)

        self.diskmapping["on_reboot"].set_active(1)

        self.diskedit.showtree()
        self.use_bootloader.set_active(False)
        self.bootloader.set_sensitive(False)

    def show(self, widget, blank=False, rerun=False):
        """Displays the settings dialog"""

        self.blank = blank

        if not rerun:
            if not blank and self.domindex.cursel.isResident:
                self.type = "memory"
                self.inmemory.set_sensitive(True)
                self.__noOpActive(self.inmemory, True)

                if self.domindex.cursel.file:
                    self.ondisk.set_sensitive(True)
                else:
                    self.ondisk.set_sensitive(False)

                self.__showMem()
            else:
                self.type = "disk"
                self.inmemory.set_sensitive(False)
                self.__noOpActive(self.ondisk, True)
                self.ondisk.set_sensitive(True)

                if blank:
                    self.__showBlank()
                else:
                    self.__showDisk()

        res = self.dialog.run() == gtk.RESPONSE_OK

        if not res:
            self.dialog.hide()
            return

        if self.type == "memory":
            domindex.cursel.setMem(int(self.memory.get_value()))
            domindex.cursel.setVCPUs(int(self.vcpu.get_value()))
        else:
            errmsgs = []

            filename = self.filename.get_text()

            if not filename:
                errmsgs.append("You must specify a file name.")

            if not self.name.get_text():
                errmsgs.append("You must specify a name for the domain.")

            if not self.use_bootloader.get_active():
                value = self.kernel.get_text()
                if not value or not os.path.exists(value):
                    errmsgs.append("You must specify a valid %s file name." \
                                   % 'kernel')
            else:
                value = self.bootloader.get_text()
                if not value:
                    errmsgs.append("You must specify a valid bootloader.")

            if errmsgs:
                errmsg = "The following errors were found:\n"
                for msg in errmsgs:
                    errmsg += " - " + msg + "\n"

                showmsg(errmsg)
                self.show(widget, blank, True)
                return

            if os.path.exists(filename) and \
                   ((not blank and filename != self.domfile.filename) \
                    or blank) and not \
                    confirmation("Do you wish to overwrite: %s" % filename):
                self.show(widget, blank, True)
                return
                    
            if self.blank:
                self.domfile = DomFile(filename)
            else:
                self.domfile.filename = filename

            # Consider enhancing the for loop below to test whether config
            # elements have actually changed and only update the DomFile
            # instance entry for those that have changed

            for entry, widget in self.diskmapping.iteritems():
                if entry in self.textfields:
                    self.domfile[entry] = widget.get_text()

                elif entry in self.valuefields:
                    self.domfile[entry] = int(widget.get_value())

                elif entry in self.combofields:
                    self.domfile[entry] = self.combomapping[widget.get_active()]

            self.domfile["disk"] = [
                repr(i) for i in self.diskedit.getDiskEntries()
                ]

            # reset depending on bootloader being used or not.
            if self.use_bootloader.get_active() and \
                   self.bootloader.get_text() and \
                   self.bootloader.get_text().strip() != '':
                self.domfile['ramdisk'] = ''
                self.domfile['kernel'] = ''
            else:
                self.domfile['bootloader']=''
                
            self.domfile.write()
            domindex.domselected()
        self.dialog.hide()

class CreateDialog:
    """ Class that handles events from create dom dialog"""

    # keep track so that signals get connected once. We may need to find
    # better solution for this.
    initialized = False 
    
    def __init__(self, wtree):
        """ Constructor"""
        self.dialog      = wtree.get_widget("CreateDialog")
        self.domName     = wtree.get_widget("DomName")
        self.domMemory   = wtree.get_widget("DomMemory")
        self.domDiskSize = wtree.get_widget("DomDiskSize")
        self.domDiskPath = wtree.get_widget("DomDiskPath")
        self.useLVM      = wtree.get_widget("UseLVM")
        self.LVGList    = wtree.get_widget("LVGList")
        self.LVGLabel   = wtree.get_widget("LVGLabel")
        
        # if no lvm then disable the checkbox
        if lvm == None:
            self.useLVM.set_sensitive(FALSE)
           
        
        # setup handlers

        if not CreateDialog.initialized:
            wtree.signal_connect("on_createDialog_cancel_button_clicked",
                                 self.on_cancel_button_clicked)
            wtree.signal_connect("on_createDialog_ok_button_clicked",
                                 self.on_ok_button_clicked)
            wtree.signal_connect("on_createDialog_DomDiskDirButton_clicked",
                                 self.on_disk_dir_button_clicked)
            wtree.signal_connect("on_createDialog_UseLVM_toggled",
                                 self.on_useLVM_toggled)
            CreateDialog.initialized = True

        self.init()
        self.show()

    def init(self):
        self.domName.set_text("")
        self.domMemory.set_text("256")
        self.domDiskSize.set_text("3")
        if (config.getDefault(prop_disks_dir) is not None):
            self.domDiskPath.set_text(config.getDefault(prop_disks_dir))
        else:
            self.domDiskPath.set_test('')
    
    def show(self):
        """ Displays the create Dom dialog"""
        self.dialog.show() 

    def on_cancel_button_clicked(self, widget):
       """  Cancel creation of dom """
       self.dialog.hide()

    def on_ok_button_clicked(self, widget):
        """  Create Dom """

        if self.domName.get_text() in domindex.domdict.keys():
            if isinstance(domindex[self.domName.get_text()], DomMem):
                showmsg("Running Domain with the same name exists.")
                return False
            
        # ask permission to overwrite if the domname
        # has already been quickcreated.
        if os.path.isfile(os.path.join(config.getDefault(prop_xenconf_dir),self.domName.get_text())):
            # domain conf file exists
            if confirmation('"%s" already exists: delete and replace?' % self.domName.get_text()):
                # permission to overwrite granted. cleanup exisiting files.
                cleanupQCDomain(self.domName.get_text(),self.domDiskPath.get_text())
            else:
                # permission not granted. abort domain creation
                showmsg('Aborting Domain Creation')
                self.dialog.hide()
                return
            
        msg = ''
        try:
            diskpath =  self.domDiskPath.get_text()
            if self.useLVM.get_active():
                diskpath = self.LVGList.child.get_text()

            quickcreateDom(self.domName.get_text(),
                           self.domMemory.get_text(),
                           diskpath,
                           int(float(self.domDiskSize.get_text())*1024),
                           self.useLVM.get_active())
        except (IOError, OSError), err:
            msg = 'Failed: '+str(err)
        except Exception, err:
            msg = 'Failed: '+str(err)
            raise
        else:
            msg = 'Creation Successful'

        showmsg(msg)
        self.dialog.hide()
        
    def on_disk_dir_button_clicked(self, widget):
        (res, dirname) = \
              file_selection("Choose location for creating file for Dom Disk",
                             "select_folder", parentwin=self.dialog)
        
        if res and dirname:
            self.domDiskPath.set_text(dirname)

    def on_useLVM_toggled(self, widget, data=None):
        if widget.get_active():
            self.domDiskPath.set_sensitive(False)
            self.LVGList.set_sensitive(True)
            self.LVGLabel.set_sensitive(True)
            self.populateLVGList()
        else:
            self.LVGList.set_sensitive(False)
            self.LVGLabel.set_sensitive(False)
            self.domDiskPath.set_sensitive(True)

    def populateLVGList(self):
        lvglist_store = gtk.ListStore(gobject.TYPE_STRING)

        textrenderer = gtk.CellRendererText()
        self.LVGList.set_model(lvglist_store)
        self.LVGList.set_text_column(0)
                
        slistmodel = gtk.TreeModelSort(lvglist_store)
        slistmodel.set_sort_column_id(0, gtk.SORT_ASCENDING)
        lvgs = lvm.listVolumeGroups()
        for lvg in lvgs:
            self.LVGList.append_text(lvg)
            
        if len(lvgs) >= 1:
            self.LVGList.set_active(0)
        

class InitParamsDialog:
    """ Class that handles events from collecting intial parameters dialog"""
    
    initialized = False
    
    def __init__(self, wtree):
        """ Constructor"""
        self.dialog = wtree.get_widget("initParamsDialog")
        self.defDiskLocation = \
                     wtree.get_widget("initParamsDialog_def_disk_location")
        self.defSnapshotLocation = \
                     wtree.get_widget("initParamsDialog_def_snapshot_location")

        if not InitParamsDialog.initialized :
            # setup handlers
            wtree.signal_connect("on_initParamsDialog_cancel_button_clicked",
                                 self.on_cancel_button_clicked)
            wtree.signal_connect("on_initParamsDialog_ok_button_clicked",
                                 self.on_ok_button_clicked)
            wtree.signal_connect("on_initParamsDialog_DiskLocationButton_clicked",
                                 self.on_disk_location_button_clicked)
            wtree.signal_connect("on_initParamsDialog_SnapshotLocationButton_clicked",
                                 self.on_snapshot_location_button_clicked)
            InitParamsDialog.initialized = True
            
        self.init()
        self.show()

    def init(self):
        self.defDiskLocation.set_text("")
        self.defSnapshotLocation.set_text("")
        
    def show(self):
        """ Displays initialization params dialog"""
        self.dialog.show() 

    def on_cancel_button_clicked(self, widget):
       """  Cancel on initialization params """
       self.dialog.hide()
       
    def on_ok_button_clicked(self, widget):
       """  Cancel creation of dom """
       config.setDefault(prop_disks_dir,
                         self.defDiskLocation.get_text())
       config.setDefault(prop_snapshots_dir,
                         self.defSnapshotLocation.get_text())
       self.dialog.hide()

    def on_disk_location_button_clicked(self, widget):
        (res, dirname) = file_selection("Select default location for Dom disks",
                                        "select_folder", parentwin=self.dialog)
        
        if res and dirname:
            self.defDiskLocation.set_text(dirname)

    def on_snapshot_location_button_clicked(self, widget):
        (res, dirname) = file_selection("Select default location for Dom snapshots",
                                        "select_folder", parentwin=self.dialog)
        
        if res and dirname:
            self.defSnapshotLocation.set_text(dirname)

class XMConfig(ConfigParser.SafeConfigParser):
    """ XenMan's configuration management class. """

    def __init__(self):
        """Looks for xenman.conf in current, user home and /etc directories. If
        it is not found, seeds one in the local directory."""

        ConfigParser.SafeConfigParser.__init__(self)

        filelist = [x for x in [os.path.join(os.getcwd(),'xenman.conf'),
                                os.path.expanduser('~/xenman.conf'),
                                '/etc/xenman.conf'] if os.path.isfile(x)]

        if len(filelist) < 1:
            print 'No Configuration File Found'
            print 'Creating default xenman.conf in current directory'

            self.configFile = os.path.join(os.getcwd(), 'xenman.conf')
            if not self.defaults(): 
                self.set('DEFAULT', prop_disks_dir, '')
                self.set('DEFAULT', prop_snapshots_dir, '')
                self.set('DEFAULT', prop_snapshot_file_ext, '.snapshot.xm')
                self.set('DEFAULT', prop_xenconf_dir,'/etc/xen')
                self.set('DEFAULT', prop_cache_dir, '/var/cache/xenman')
                self.set('DEFAULT', prop_kernel, '')
                self.set('DEFAULT', prop_ramdisk, '')
                self.set('DEFAULT', prop_bootloader, '/usr/bin/pygrub')
                self.set('DEFAULT', prop_staging_location, '')
                self.set('DEFAULT', prop_lvm, 'True')
                self.__commit()
        else:
            print 'Configuration File found.\nUsing: %s' % filelist[0]
            self.configFile = filelist[0]
            self.read(self.configFile)

        self.set('DEFAULT',prop_dom0_kernel,platform.release())
        self.__commit()

    def __commit(self):
        outfile = open(self.configFile,'w')
        self.write(outfile)
        outfile.close()

    def getDefault(self, option):
        """ retrieve a default option/key value """

        # does the option exist? return None if not
        if not self.has_option('DEFAULT', option): return None

        # option is available in the config. get it.
        retVal = self.get('DEFAULT', option)
        
        # check if the value is blank. if so, return None
        # otherwise, return the value.
        if not retVal.strip():
            return None
        else:
            return retVal

    def setDefault(self, option, value):
        """set the default for option to value.
        POSTCONDITION: option, value pair has been set in the default
        configuration, and committed to disk"""

        self.set('DEFAULT', option, value)
        self.__commit()


class LVMProxy:
    """A thin, (os-dependent) wrapper around the shell's LVM
    (Logical Volume Management) verbs"""
    # TODO: move this class to an OSD module
    
    def __init__(self):
        """ The constructor simply checks if LVM services are available
        for use via the shell.
        RAISES: OSError"""
        self.call_prefix = ''
        if subprocess.call(['vgs 2> /dev/null'], shell=True, close_fds=True):
            self.call_prefix = '/usr/sbin/'
            if subprocess.call(['/usr/sbin/vgs 2> /dev/null'], shell=True, close_fds=True):
                raise OSError("LVM facilities not found")

    def listVolumeGroups(self):
        """ Returns the list of existing Volume Groups"""
        try:
            vglist = subprocess.Popen([self.call_prefix + 'vgs -o vg_name --noheadings'],stdout=subprocess.PIPE, \
                                      universal_newlines=True, shell=True, \
                                      close_fds=True).communicate()[0]
            return [s.strip() for s in vglist.splitlines()]
        except OSError, err:
            print err
            return None

    def listLogicalVolumes(self, vgname):
        """ Returns the list of Logical Volumes in a Volume Group"""
        try:
            lvlist = subprocess.Popen([self.call_prefix + 'lvs -o lv_name --noheadings '+ vgname],stdout=subprocess.PIPE, \
                                      universal_newlines=True, shell=True, \
                                      close_fds=True).communicate()[0]
            return [s.strip() for s in lvlist.splitlines()]
        except OSError, err:
            print err
            return None

    def createLogicalVolume(self, lvname, lvsize, vgname):
        """ Create a new LV with in the specified Volume Group.
        'lvsize' denotes size in number of megabytes.
        RETURNS: True on sucees
        RAISES: OSError"""
        p1 = subprocess.Popen(self.call_prefix + 'lvcreate %s -L %sM -n %s'%(vgname,lvsize,lvname),\
                              stderr=subprocess.PIPE,universal_newlines=True, \
                              shell=True, close_fds=True)
        error = p1.communicate()[1]
        if (p1.returncode):
            raise OSError(error.strip('\n'))
        else:
            return True
        
                
    def removeLogicalVolume(self, lvname, vgname=None):
        """ Remove the logical volume 'lvname' from the
        Volume Group 'vgname'. If the latter is not specified,
        'lvname' is treated as a fully specified path
        RETURNS: True on success
        RAISES: OSError"""
        if (vgname):
            lvpath = vgname + '/' + lvname
        else:
            lvpath = lvname
            
        p1 = subprocess.Popen(self.call_prefix + 'lvremove -f %s'% lvpath, \
                              stderr=subprocess.PIPE,universal_newlines=True, \
                              shell=True, close_fds=True)
        error = p1.communicate()[1]
        if (p1.returncode):
            raise OSError(error.strip('\n'))
        else:
            return True


#
# function definitions
#

def quitProgram(*args):
    """Quit the execution of the program"""

    gtk.main_quit()

def show_about(*args):
    """handler that shows the about dialog"""

    wtree.get_widget('AboutDialog').show()

def hide_about(widget, *args):
    """handler that hides the about dialog"""

    widget.hide()
    return True

def showmsg(msg):
    """Displays the status message passed as an argument in a popup window"""

    wtree.get_widget('StatusMsg').set_text(msg)
    dialog = wtree.get_widget('StatusDialog')
    dialog.run()
    dialog.hide()

def confirmation(msg):
    """Displays a confirmation dialog message and returns a boolean
    representing the result"""

    wtree.get_widget('ConfirmationDialog').set_title("Confirm")
    wtree.get_widget('ConfirmationMsg').set_text(msg)
    dialog = wtree.get_widget('ConfirmationDialog')
    res = dialog.run() == gtk.RESPONSE_YES
    dialog.hide()
    return res

def file_selection(title, fstype, init_folder=None, init_file=None,
                   parentwin=None):
    """This function will setup a file selection dialogue and return a tuple
    consisting of a boolean which is true if the user selected a filename,
    followed by the filename. fstype is either "open" or "save" or
    "select_folder" and title is whateveryou want displayed as the
    window's title bar. The init_folder and init_file params would be used
    as initial values for the dialog box"""

    if fstype == "save":
        filesel = gtk.FileChooserDialog(title,
                                        action=gtk.FILE_CHOOSER_ACTION_SAVE,
                                        buttons=(gtk.STOCK_CANCEL,
                                                 gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_SAVE,
                                                 gtk.RESPONSE_OK))
    elif fstype == "open":
        filesel = gtk.FileChooserDialog(title,
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,
                                                 gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
    elif fstype == "select_folder":
        filesel = gtk.FileChooserDialog(title,
                                        action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                        buttons=(gtk.STOCK_CANCEL,
                                                 gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
    if init_folder is not None:
        filesel.set_current_folder(init_folder)
    if init_file is not None:
        filesel.set_current_name(init_file)
    
    

    if not parentwin:
        parentwin = wtree.get_widget('MainWindow')
    filesel.set_transient_for(parentwin)
    filesel.set_position(gtk.WIN_POS_CENTER_ON_PARENT)

    res = filesel.run() == gtk.RESPONSE_OK

    if res:
        filename = filesel.get_filename()
    else:
        filename = None

    filesel.destroy()

    return (res, filename)

def savedom(widget):
    """Take Snapshot of a running dom. (a.k.a save)"""

    if domindex.cursel.isDomU():
        snap_dir = config.getDefault(prop_snapshots_dir)
        default_ext = config.getDefault(prop_snapshot_file_ext);
        snap_file = domindex.cursel.name
        if default_ext is not None:
            snap_file += default_ext
        
        (res, fname) = file_selection("Save Snapshot", "save", snap_dir,
                                      snap_file)
        if res and fname:
            domindex.cursel.save(fname)

def restoredom(widget):
    """Restore a dom from a file."""

    snap_dir = config.getDefault(prop_snapshots_dir)
    (res, fname) = file_selection("Restore from Snapshot","open", snap_dir)
    if res and fname:
        server.xend.domain.restore(fname)
        domindex.refresh()

def rebootdom(widget):
    """Reboot the currently selected domain"""

    if domindex.cursel.isDomU():
        if confirmation("Reboot " + domindex.cursel.name + "?"):
            domindex.cursel.reboot()

def shutdowndom(widget):
    """shutdown the currently selected domain"""

    if domindex.cursel.isDomU():
        if confirmation("Shutdown " + domindex.cursel.name + "?"):
            domindex.cursel.shutdown()

def startdom(widget):
    """Start the selected domain"""

    if not domindex.cursel.isResident:
        errmsgs = domindex.cursel.validate()

        if errmsgs:
            errmsg = "Can't start the domain because of the following " + \
                     "problems:\n"

            for msg in errmsgs:
                errmsg += " - " + msg + "\n"

            showmsg(errmsg)
            return

        try:
            domindex.cursel = domindex.cursel.start()
        except OSError ,e:
            showmsg("Error starting " + domindex.cursel.name +". "+ str(e))
            return
        domindex.refresh()
        domindex.notebook.showTab(domindex.cursel)
        loop.skipcount = 1
        loop.uirefresh = True

def killdom(widget):
    """Kills the currently selected domain"""

    if domindex.cursel.isDomU():
        if (confirmation("Kill " + domindex.cursel.name + "?")):
            # sometime the xen destroy gives socket error after destorying vm
            # lets ignore it for now.
            try:
                domindex.cursel.destroy()
                domindex.refresh()
            except (socket.error, xmlrpclib.Fault):
                pass

            domindex.cursel = domindex[0]
            loop.uirefresh = True

def pausedom(widget):
    """Pause the currently selected domain"""

    if domindex.cursel.isDomU() and domindex.userpause:
        state = domindex.cursel["state"]
        if state[2] == 'p':
            if confirmation("Resume " + domindex.cursel.name + "?"):
                domindex.cursel.resume()
            domindex.stateRefresh()
        else:
            if confirmation("Pause " + domindex.cursel.name + "?"):
                domindex.cursel.pause()
            domindex.stateRefresh()
                
def getNodeVal(parm):
    """Retrieve a node wide parameter"""

    return DomMem.searchtree(server.xend.node.info(), parm)

def popupmenu(widget, event):
    """Checks whether the right mouse button was clicked within the domain
    TreeView and if so, displays a popup menu"""

    if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
        wtree.get_widget("PopupMenu").popup(None, None, None, event.button,
                                            event.time)

def opendom(widget):
    """Handles the open domain menu item which adds a dom by file name to
    the list of doms in the UI"""

    result, filename = \
            file_selection("Please select a Dom configuration file", "open",
                           config.getDefault(prop_xenconf_dir))

    if result and filename:
        domindex.addDomFile(filename)

def removedom(widget):
    """Handles the menu items requesting to removal of a dom from the
    domindex"""
    cleanupQCDomain(domindex.cursel.name)
    domindex.remDomFile(domindex.cursel.filename)

def show_create_dialog(widget):
    CreateDialog(wtree)

def assertPreconditions():
    """Checks a number of required pre-conditions for the tool to be usable,
    if those aren't met, displays an appropriate dialog box and exits"""

    ts = rpm.TransactionSet()

    fedora = ts.dbMatch(rpm.RPMTAG_NAME, "fedora-release").count()

    # Check that we're running as root
    if os.getuid() != 0:
        showmsg("You must be running as root to interact with the hypervisor")
        sys.exit(1)

    # Check that we're running a Dom0 Xen kernel
    if platform.release().find('xen') == -1:
        msg = "You must be running a Xen Dom0 kernel to manage Xen domains."

        if fedora:
            if ts.dbMatch(rpm.RPMTAG_NAME, "kernel-xen0").count():
                msg += "\nYou seem to have an appropriate Xen Dom0 kernel " + \
                       "installed, so restart your system and boot into it."
            else:
                msg += "\nYou don't seem to have an appropriate kernel " + \
                       'installed, try running "yum install "'

                if not ts.dbMatch(rpm.RPMTAG_NAME, "xen").count():
                    msg += "xen "

                msg += 'kernel-xen0" to deploy the appropriate kernel, ' + \
                       "then restart your system and boot into that newly " + \
                       "installed kernel."

        showmsg(msg)
        sys.exit(1)
        
    try:
        server.xend.domains()
    except (XendError, xmlrpclib.Fault):
        if not ts.dbMatch(rpm.RPMTAG_NAME, "xen").count():
            msg = "You must install the Xen daemon in order manage Xen " + \
                  "domains."
            if fedora:
                msg += 'Try running "yum install xen", followed by ' + \
                       '"service xend start" as root.'
        else:
            msg = "The Xen daemon isn't responding, "

            if fedora:
                msg += 'try starting it using "service xend start" as root.'
            else:
                msg += "please ensure it is installed and started."

        showmsg(msg)
        sys.exit(1)

def fetchImage(src, dest):
    """ Copies 'src' to 'dest'. 'src' can be an http or ftp URL
    or a filesystem location. dest must be a fully qualified
    filename."""
    
    print "Fetching: "+src
    if src.startswith("http://") or src.startswith("ftp://"):
        # newtwork fetch
        urllib.urlretrieve(src,dest)
    else:
        # filesystem fetch
        shutil.copyfile(src, dest)

def validateConfig():
    """Checks for existance of all critical paths and
    files specified in the DEFAULT's section the config file.
    Creates directories or fetches kernel/ramdisk from staging
    location if necessary.
    RETURNS: False if validation could not succeed"""
 
    # check xenman cache directory
    if config.getDefault(prop_cache_dir):
        if not os.path.exists(config.getDefault(prop_cache_dir)):
            try:
                os.mkdir(config.getDefault(prop_cache_dir))
            except OSError:
                showmsg('ERROR: Cannot create default cache directory')
                return False
    else:
        showmsg("Invalid Configuration File: 'cache_dir' not set")
        return False


    # check default kernel
    if not config.getDefault(prop_kernel):
        # default kernel not set. get it from the staging area
        kernel = os.path.join(config.getDefault(prop_cache_dir),
                              'vmlinuz.default')
        try:
            fetchImage(config.getDefault(prop_staging_path_kernel),
                       kernel)
            # set the fetched image as the default kernel
            config.setDefault(prop_kernel,kernel)
        except:
            showmsg('Invalid Staging Area: cannot fetch kernel image')
            return False


    # check default ramdisk
    if not config.getDefault(prop_ramdisk):
        # default ramdisk not set. get it from the staging area
        ramdisk = os.path.join(config.getDefault(prop_cache_dir),
                               'initrd.img.default')
        try:
            fetchImage(config.getDefault(prop_staging_path_ramdisk),
                       ramdisk)
            # set the fetched image as the default ramdisk
            config.setDefault(prop_ramdisk, ramdisk)
        except:
            showmsg('Invalid Staging Area: cannot fetch ramdisk image')
            return False

    return True

def createVBD(filename, size=2048, backup=False):
    """Creates a file Virtual Block Device called 'name'
    at 'location'. 'size' is specified in kB.
    RAISES: IOError, OSError """

    if os.path.isfile(filename):
        if backup:
            os.rename(filename,filename+'.bak')
        else:
            os.remove(filename)

    # create the new disk block
    # - adapted directly from xenguest-install
    fd = os.open(filename, os.O_WRONLY | os.O_CREAT)
    off = size * 1024L**2
    os.lseek(fd, off, 0)
    os.write(fd, '\x00')
    os.close(fd)

def quickcreateDom(name, mem, vbdlocation, vbdsize=2048, uselvm=False):
    """Creates a new dom configuration file using default
    kernel and ramdisk attributes and starts the corresponding
    DomU instance.
    RAISES: IOError, OSError"""

    # check that everything is in order
    if not validateConfig():
        raise "Config not valid"

    # create a new domfile object
    domfile = DomFile(os.path.join(config.getDefault(prop_xenconf_dir),name))
 
    # set domfile attributes
    domfile['name'] = name
    domfile['kernel'] = config.getDefault(prop_kernel)
    domfile['ramdisk'] = config.getDefault(prop_ramdisk)
    domfile['memory'] = int(mem)
    domfile['vif'].append('')
    domfile['on_reboot'] = 'destroy'
    domfile['on_crash'] = 'destroy'

    if not uselvm :
        # create VBD and set corresponding attribute in domfile
        filename = "%s/%s.disk.xm" % (vbdlocation, name)
        createVBD(filename, vbdsize)
        domfile['disk'].append('file:%s,hda,w' % filename)
    else:
        # create a new logical volume in the 'vbdlocation' volume group.
        if(lvm.createLogicalVolume(name+'.disk.xm',vbdsize,  vbdlocation)):
            domfile['disk'].append('phy:/dev/%s/%s.disk.xm,hda,w'%(vbdlocation,name))
        else:
            raise OSError('could not create logical volume')

    # commit domfile to disk and start.w/ default ramdisk
    # TODO: avoid writing the domfile here and try to do the
    #       first launch w/ kernel & ramdisk passed in
    #       directly.
    domfile.write()
    newdom = domfile.start()
    domindex.refresh()
    domindex.cursel = newdom
    domindex.notebook.showTab(newdom)
    domindex.addDomFile(domfile.filename)

    # alter domfile to use indigenous ramdisk on
    # subsequent boots.
    # TODO: integrate this into the stanza above.
    domfile['kernel'] = ''
    domfile['ramdisk']= ''
    domfile['bootloader'] = config.getDefault(prop_bootloader)
    domfile.write()

def cleanupQCDomain(name, vbdlocation=''):
    """ Delete the xen configuration file and associated
    disk devices created during a quickcreate"""
    domfilename = os.path.join(config.getDefault(prop_xenconf_dir),name)
    if os.path.isfile(domfilename):
        os.remove(domfilename)
        
    if name in domindex.domdict.keys():
        for file in domindex[name].getDisks():
            if file.type is 'lvm':
                lvm.removeLogicalVolume(file.filename)
                print 'deleting: ' + file.filename
            elif file.type is 'file':
                if os.path.isfile(file.filename):
                    print 'deleting: ' + file.filename
                    os.remove(file.filename)
    else:
        print "Couldn't find the domain %s. Skipping deletion" % name
                
def getServer():
    """ Look at the arguments and return the object to connect to
    either remote or local server"""
    count = len(sys.argv) 
    hostname = None
    protocol = None
    if count > 1:
        hostname = sys.argv[1]
    if count > 2:
        protocol = sys.argv[2]
            
    if protocol is not None:
        if protocol not in ["tcp", "ssh"]: 
            print "Invalid protocol specified"
            sys.exit(1)
                    
    if hostname == None and protocol == None:
        server = defaultServer
    else:
        if hostname is not None and protocol == None:
            protocol = "tcp"

        if protocol == "tcp":
            server = ServerProxy("http://" + hostname + ":" + \
                                 default_tcp_port + default_path)
                                
        if protocol == "ssh":
            server = ServerProxy("ssh://" + default_user + "@" + \
                                 hostname + default_path)
            
    return server   
                

        
#
# Main block
#


# Look at the command line arguments  and set up the server.
# for now simple checks for arguments

server = getServer()
wtree = gtk.glade.XML(gladefile)

# Sanity check the environment before performing any Xen interactions

assertPreconditions()


# Initialise the configuration object to gain access to global defaults
 
config = XMConfig()          # global configuration

# Let's instantiate the dom management object

domindex = DomIndex(wtree)

settings = DomSettings(wtree, domindex)

signals = {"on_MainWindow_delete_event": quitProgram,
           "on_startbutton_clicked": startdom,
           "on_killbutton_clicked": killdom,
           "on_pausebutton_toggled": pausedom,
           "on_new_domain_activate": (settings.show, True),
           "on_open_domain_activate": opendom,
           "on_Remove_activate": removedom,
           "on_create_domain_wizard_activate": show_create_dialog,
           "on_DomView_button_press_event": popupmenu,
           "on_Settings_activate": settings.show,
           "on_quit_activate": quitProgram,
           "on_rebootbutton_clicked": rebootdom,
           "on_shutdownbutton_clicked": shutdowndom,
           "on_snapshotbutton_clicked": savedom,
           "on_restorebutton_clicked": restoredom,
           'on_about_activate': show_about,
           'on_AboutDialog_response': hide_about,
           'on_AboutDialog_delete_event': hide_about,
           }

wtree.signal_autoconnect(signals)

# Let's define the parent for the dialog windows (glade should do it but
# doesn't)

mainwin = wtree.get_widget('MainWindow')

for dname in ("SettingsDialog", "StatusDialog", "ConfirmationDialog",
              "AboutDialog", "CreateDialog", "initParamsDialog"):
    dialog = wtree.get_widget(dname)
    dialog.set_transient_for(mainwin)


# Initialise the LVM proxy in preperation for interaction with
# Logical Volume Management facilities (Linux only)

lvm = None
if (config.getDefault(prop_lvm).lower() == 'true'):
    lvm = LVMProxy()

# get default  parameters for disks and snapshot location
if config.getDefault(prop_disks_dir) == None or \
       config.getDefault(prop_snapshots_dir) == None :
    InitParamsDialog(wtree)

# Setup the refresh loop

loop = RefreshLoop(wtree, domindex)


# Run the main GUI loop

gtk.main()
